<?php
// $Id: sphinxsearch.taxonomy.inc,v 1.1.2.2 2008/08/28 04:49:36 markuspetrux Exp $

/**
 * @file
 * Common taxonomy related functions for the Sphinx search module.
 */

/**
 * Obtain list of enabled vocabularies.
 *
 * Each vocabulary will be indexed on a separate multi-valued field.
 *
 * @return array
 *   list of enabled vocabularies.
 */
function sphinxsearch_get_enabled_vocabularies() {
  static $vocabularies;
  if (!isset($vocabularies)) {
    $vocabularies = array();
    if (module_exists('taxonomy')) {
      foreach(taxonomy_get_vocabularies() as $vocabulary) {
        if (variable_get('sphinxsearch_include_vocabulary_'. $vocabulary->vid, 0)) {
          $vocabularies[(int)$vocabulary->vid] = $vocabulary;
        }
      }
    }
  }
  return $vocabularies;
}

/**
 * Encode terms hash into a comma separated list.
 *
 * @param array $terms
 *   Hash of terms.
 * @param string $separator
 *   Terms separator. Defaults to ', '.
 * @return string
 *   Comma separated list of terms.
 */
function sphinxsearch_taxonomy_encode_typed_terms($terms, $separator = ', ') {
  $typed_terms = array();
  foreach ($terms as $tid => $term) {
    // Commas and quotes in terms are special cases, so encode 'em.
    if (strpos($term->name, ',') !== FALSE || strpos($term->name, '"') !== FALSE) {
      $term->name = '"'. str_replace('"', '""', $term->name) .'"';
    }
    $typed_terms[] = $term->name;
  }
  return implode($separator, $typed_terms);
}

/**
 * Decode a comma separated list of strings into hash of terms.
 * @see taxonomy_node_save()
 *
 * @param int $vid
 *   Vocabulary identifier related to terms.
 * @param string $typed_terms
 *   Comma separated list of terms.
 * @return array
 *   Hash of terms. Note that a string of not found typed terms is stored as tid -1.
 */
function sphinxsearch_taxonomy_decode_typed_terms($vid, $typed_terms) {
  $terms = array();
  $regexp = '%(?:^|,\ *)("(?>[^"]*)(?>""[^"]* )*"|(?: [^",]*))%x';
  preg_match_all($regexp, $typed_terms, $matches);
  $typed_terms = array_unique($matches[1]);
  foreach ($typed_terms as $typed_term) {
    $typed_term = trim(str_replace('""', '"', preg_replace('#^"(.*)"$#', '\1', $typed_term)));
    if (!empty($typed_term)) {
      $possibilities = taxonomy_get_term_by_name($typed_term);
      $term = NULL;
      foreach ($possibilities as $possibility) {
        if ($possibility->vid == $vid) {
          $term = $possibility;
        }
      }
      if ($term) {
        $terms[$term->tid] = $term;
      }
      else {
        // Store typed terms that we haven't found in a particular item with tid = -1,
        // so caller has the chance to notify the user about it.
        if (!isset($terms[-1])) {
          $terms[-1] = array();
        }
        $terms[-1][] = $typed_term;
      }
    }
  }
  if (isset($terms[-1])) {
    $terms[-1] = implode(', ', $terms[-1]);
  }
  return $terms;
}

/**
 * Get term data for the given tids.
 *
 * @param int $vid
 *   Vocabulary identifier related to terms.
 * @param array $tids
 *   List of tids.
 * @return array
 *   Hash of terms. Note that a string of not found tids is stored as tid -1.
 */
function sphinxsearch_taxonomy_get_terms($vid, $tids) {
  $terms = array();
  if (!empty($tids)) {
    $tids_count = count($tids);
    $query_args = array_merge(array($vid), $tids);
    $query_sql = 'SELECT t.* FROM {term_data} t WHERE t.vid = %d AND t.tid ';
    if ($tids_count == 1) {
      $query_sql .= '= %d';
    }
    else {
      $placeholders = implode(',', array_fill(0, count($tids), '%d'));
      $query_sql .= 'IN ('. $placeholders .')';
    }
    $result = db_query(db_rewrite_sql($query_sql, 't', 'tid'), $query_args);
    while ($term = db_fetch_object($result)) {
      $terms[$term->tid] = $term;
    }
    $diff = array_diff($tids, array_keys($terms));
    if (!empty($diff)) {
      $terms[-1] = implode(', ', $diff);
    }
  }
  return $terms;
}

/**
 * Build tagadelic items for the given Sphinx search results.
 *
 * @param array $search_results
 *   Search results structure.
 * @param int $vid
 *   Vocabulary ID.
 * @param int $levels
 *   Amount of tag-sizes.
 * @param string $sortmode
 *   Sort order options.
 *
 * @return array
 */
function sphinxsearch_tagadelic_build_tags($search_results, $vid, $levels = 10, $sortmode = 'title,asc') {
  $tags = array();
  if ($search_results['total_available'] > 0) {
    $terms = sphinxsearch_taxonomy_get_terms($vid, array_keys($search_results['groups']));

    // Find minimum and maximum log-count. Algorithm based on tagadelic.module.
    $min = 1e9;
    $max = -1e9;
    foreach ($search_results['groups'] as $tid => $group_info) {
      if (isset($terms[$tid])) {
        $tag = $terms[$tid];
        $tag->tagadelic_count = log($group_info['count']);
        $min = min($min, $tag->tagadelic_count);
        $max = max($max, $tag->tagadelic_count);
        $tags[$tid] = $tag;
      }
    }
    unset($terms);

    // Note: we need to ensure the range is slightly too large to make sure even
    // the largest element is rounded down.
    $range = max(.01, $max - $min) * 1.0001;
    foreach ($tags as $tid => $tag) {
      $tags[$tid]->tagadelic_weight = 1 + floor($levels * ($tag->tagadelic_count - $min) / $range);
    }

    // Sort tags.
    list($sort_by, $sort_order) = explode(',', $sortmode);
    switch ($sort_by) {
      case 'title':
        usort($tags, create_function('$a,$b', 'return strnatcasecmp($a->name, $b->name);'));
        break;
      case 'weight':
        usort($tags, create_function('$a,$b', 'return $a->tagadelic_weight > $b->tagadelic_weight;'));
        break;
      case 'random':
        shuffle($tags);
        break;
    }
    if ($sort_order == 'desc') {
      $tags = array_reverse($tags);
    }
  }
  return $tags;
}

/**
 * Tagadelic block (weighted tags cloud).
 *
 * @param int $vid
 *   Vocabulary ID.
 * @param int $count
 *   Number of tags to render.
 * @param int $levels
 *   Amount of tag-sizes.
 * @param string $sortmode
 *   Sort order options.
 * @param boolean $faceted
 *   Use TRUE to build tag links for faceted search (internal use only).
 *   @see sphinxsearch_faceted_block()
 * @param array $request_options
 *   Requested search options (optional).
 *
 * @return string
 *   Themed HTML block.
 */
function sphinxsearch_tagadelic_block($vid, $count, $levels, $sortmode, $faceted = FALSE, $request_options = array()) {
  // Build search options structure.
  $search_options = sphinxsearch_parse_request($request_options);
  if (!isset($search_options['filters']['taxonomy'])) {
    $search_options['filters']['taxonomy'] = array($vid => array());
  }
  else if (!isset($search_options['filters']['taxonomy'][$vid])) {
    $search_options['filters']['taxonomy'][$vid] = array();
  }

  // Results count should include number of terms specified in filters.
  $count += count($search_options['filters']['taxonomy'][$vid]);

  // Execute a GroupBy query based on current search options.
  $search_results = sphinxsearch_execute_query(array_merge($search_options, array(
    'results_per_page' => $count,
    'group_by' => 'terms'. $vid,
  )));

  // Remove filtered terms from results.
  $tagadelic_groups = array();
  foreach ($search_results['groups'] as $tid => $group_info) {
    if (!isset($search_options['filters']['taxonomy'][$vid][$tid])) {
      $tagadelic_groups[$tid] = $group_info;
    }
  }
  $search_results['groups'] = $tagadelic_groups;

  // Compute filtered terms list in user supplied order.
  $filtered_terms = array();
  foreach ($search_options['filters']['taxonomy'][$vid] as $tid => $term) {
    if (!isset($tagadelic_groups[$tid])) {
      $filtered_terms[$tid] = $term;
    }
  }

  // Transform search results into tagadelic items.
  $tags = sphinxsearch_tagadelic_build_tags($search_results, $vid, $levels, $sortmode);

  // Additional faceted search processing (internal use only).
  // @see sphinxsearch_faceted_block()
  if ($faceted) {
    // Return weighted tags, filtered terms and search options.
    return array($tags, $filtered_terms, $search_options);
  }

  return theme('sphinxsearch_tagadelic_block', $tags);
}

/**
 * Render tagadelic block for the given tags.
 *
 * @param array $tags
 *
 * @ingroup themeable
 */
function theme_sphinxsearch_tagadelic_block($tags) {
  $output = '';

  foreach ($tags as $tag) {
    $attributes = array('class' => 'tagadelic-level'. $tag->tagadelic_weight, 'rel' => 'tag');
    $url = (!empty($tag->faceted_path) ? url($tag->faceted_path, $tag->faceted_query) : url(taxonomy_term_path($tag)));
    // Note: we don't use l() because it may append class="active"
    // and this breaks the look of our pretty nice clouds.
    $output .= '<a href="'. $url .'"'. drupal_attributes($attributes) .'>'. check_plain($tag->name) .'</a>' ."\n";
  }

  if (empty($output)) {
    return '';
  }
  return '<div class="tagadelic-wrapper">'. $output .'</div>';
}

/**
 * Render more link for tagadelic block.
 *
 * @param int $vid
 *
 * @ingroup themeable
 */
function theme_sphinxsearch_tagadelic_more($vid) {
  return '<div class="more-link">'. l(t('more'), 'tagadelic/'. $vid) .'</div>';
}

/**
 * Menu callback; Tagadelic page (weighted tags cloud).
 *
 * @param string $vids
 */
function sphinxsearch_tagadelic_page($vids = NULL) {
  $vocabularies = sphinxsearch_get_enabled_vocabularies();
  if (!isset($vids)) {
    $vids = array_keys($vocabularies);
  }
  else {
    $vids = array_filter(array_map('intval', array_map('trim', explode(',', $vids))));
    $vids = array_intersect(array_keys($vocabularies), array_unique($vids));
  }
  $blocks = array();
  foreach ($vids as $vid) {
    $content = sphinxsearch_tagadelic_block($vid,
      (int)variable_get('sphinxsearch_page_tagadelic_tags', 100),
      (int)variable_get('sphinxsearch_page_tagadelic_levels', 10),
      variable_get('sphinxsearch_page_tagadelic_sortmode', 'title,asc')
    );
    if (!empty($content)) {
      $blocks[$vid] = array(
        'vocabulary' => $vocabularies[$vid],
        'content' => $content,
      );
    }
  }
  if (empty($blocks)) {
    drupal_not_found();
    return;
  }
  return theme('sphinxsearch_tagadelic_page', $blocks);
}

/**
 * Render tagadelic page for the given blocks.
 *
 * @param array $blocks
 *   Hash of blocks per vocabulary. Elements:
 *     - vocabulary: object
 *     - content: Tagadelic block content in HTML format.
 *
 * @ingroup themeable
 */
function theme_sphinxsearch_tagadelic_page($blocks) {
  $output = '';
  $blocks_count = count($blocks);
  foreach ($blocks as $vid => $block) {
    $output .= '<div class="tagadelic-page-block">';
    if ($blocks_count == 1) {
      drupal_set_title(t('Tags in @vocabulary', array('@vocabulary' => $block['vocabulary']->name)));
    }
    else {
      $output .= '<h2>'. check_plain($block['vocabulary']->name) .'</h2>';
    }
    $output .= $block['content'] .'</div>';
  }
  return $output;
}

/**
 * Faceted search block.
 *
 * @param array $vocabularies
 *   Hash of vocabularies to include in the block.
 * @param int $count
 *   Number of tags to render per vocabulary.
 * @param int $levels
 *   Amount of tag-sizes.
 * @param string $sortmode
 *   Sort order options.
 *
 * @return string
 *   Themed HTML block.
 */
function sphinxsearch_faceted_block($vocabularies, $count, $levels, $sortmode) {
  foreach ($vocabularies as $vid => $vocabulary) {
    // Obtain raw weighted tags to help us build faceted navigation.
    list($tags, $filtered_terms, $search_options) = sphinxsearch_tagadelic_block($vid, $count, $levels, $sortmode, TRUE, $_GET);

    // Compute faceted search path/query string for each tag.
    if (!empty($tags)) {
      foreach ($tags as $i => $tag) {
        $faceted_search_options = $search_options;
        if (!isset($faceted_search_options['filters']['taxonomy'][$vid][$tag->tid])) {
          $faceted_search_options['filters']['taxonomy'][$vid][$tag->tid] = $tag;
        }
        $tags[$i]->faceted_path = sphinxsearch_get_search_path();
        $tags[$i]->faceted_query = sphinxsearch_get_query_string($faceted_search_options);
      }
      $vocabularies[$vid]->faceted_tags = $tags;
    }

    // Compute faceted search path/query string for each term and the vocabulary itself.
    if (!empty($filtered_terms)) {
      $reversed_terms = array_reverse($filtered_terms, TRUE);
      $remaining_terms = $filtered_terms;
      foreach ($reversed_terms as $tid => $term) {
        $faceted_search_options = $search_options;
        $faceted_search_options['filters']['taxonomy'][$vid] = $remaining_terms;
        $filtered_terms[$tid]->faceted_path = sphinxsearch_get_search_path();
        $filtered_terms[$tid]->faceted_query = sphinxsearch_get_query_string($faceted_search_options);
        unset($remaining_terms[$tid]);
      }
      $faceted_search_options = $search_options;
      unset($faceted_search_options['filters']['taxonomy'][$vid]);
      $vocabularies[$vid]->faceted_path = sphinxsearch_get_search_path();
      $vocabularies[$vid]->faceted_query = sphinxsearch_get_query_string($faceted_search_options);
      $vocabularies[$vid]->faceted_terms = $filtered_terms;
    }
  }
  return theme('sphinxsearch_faceted_block', $vocabularies);
}

/**
 * Render faceted search block for the given vocabularies.
 *
 * @param array $vocabularies
 *
 * @ingroup themeable
 */
function theme_sphinxsearch_faceted_block($vocabularies) {
  $output = '';

  // Build superset section.
  $superset = array();
  foreach ($vocabularies as $vid => $vocabulary) {
    if (!empty($vocabulary->faceted_terms)) {
      $links = array();
      $last_term = array_pop($vocabulary->faceted_terms);
      foreach ($vocabulary->faceted_terms as $tid => $term) {
        $url = url($term->faceted_path, $term->faceted_query);
        $link = '<span class="faceted-superset-link"><a href="'. $url .'">'. check_plain($term->name) .'</a></span>';
        $links[] = $link;
      }
      $links[] = '<span class="faceted-superset-youarehere">'. check_plain($last_term->name) .'</span>';
      if (!empty($links)) {
        $url = url($vocabulary->faceted_path, $vocabulary->faceted_query);
        $link = '<strong class="faceted-superset-link"><a href="'. $url .'">'. check_plain($vocabulary->name) .'</a></strong>';
        $superset[] = '<div>'. implode(' › ', array_merge(array($link), $links)) .'</div>';
      }
    }
  }
  if (!empty($superset)) {
    $output .= '<div class="faceted-superset"><h2>'. t('You are here:') .'</h2><div class="faceted-map">'. theme('item_list', $superset) .'</div></div>';
  }

  // Build refine search by terms section.
  foreach ($vocabularies as $vid => $vocabulary) {
    if (!empty($vocabulary->faceted_tags)) {
      $tagadelic = theme('sphinxsearch_tagadelic_block', $vocabulary->faceted_tags);
      if (!empty($tagadelic)) {
        $output .= '<div class="faceted-vocabulary"><h2>'. t('Refine search by @vocabulary', array('@vocabulary' => $vocabulary->name)) .'</h2>'. $tagadelic .'</div>' ."\n";
      }
    }
  }

  if (empty($output)) {
    return '';
  }
  return '<div class="faceted-wrapper">'. $output .'</div>';
}
